import xml.etree.ElementTree as etree
from collections import OrderedDict
from dataclasses import dataclass
from typing import Dict

from markdown import Markdown
from markdown.extensions import footnotes
from markdown.inlinepatterns import InlineProcessor


@dataclass
class Footnote:
    text: str
    index: int
    inserted: bool = False


class FootnoteInlineProcessor(InlineProcessor):
    def __init__(self, pattern, footnotes):
        super().__init__(pattern)
        self.footnotes: FootnoteExtension = footnotes

    def handleMatch(self, m, data):
        note_id = m.group(1)
        try:
            footnote = self.footnotes.footnotes[note_id]

        except KeyError:
            return None, None, None

        else:
            if not footnote.inserted:
                el = etree.Element("span", attrib={"class": "footnote", "data-footnote-id": note_id})
                el.text = footnote.text
                footnote.inserted = True

            else:
                el = etree.Element("sup", attrib={"class": "footnote-ref", "data-footnote-id": note_id})
                el.text = str(footnote.index)

            return el, m.start(0), m.end(0)


class FootnoteExtension(footnotes.FootnoteExtension):
    def __init__(self):
        self.footnotes: OrderedDict[str, Footnote] = OrderedDict()
        self.unique_prefix = 0
        self.found_refs: dict[str, int] = {}
        self.used_refs: set[str] = set()

    def setFootnote(self, id, text):
        self.footnotes[id] = Footnote(text, len(self.footnotes) + 1)

    def extendMarkdown(self, md: Markdown):
        md.registerExtension(self)
        self.parser = md.parser
        self.md = md

        md.parser.blockprocessors.register(footnotes.FootnoteBlockProcessor(self), 'footnote', 17)
        FOOTNOTE_RE = r'\[\^([^\]]*)\]'  # blah blah [^1] blah
        md.inlinePatterns.register(FootnoteInlineProcessor(FOOTNOTE_RE, self), 'footnote', 170)



import base64
import json
import os
from io import StringIO
from subprocess import CalledProcessError, check_call
from tempfile import TemporaryDirectory

from markdown import Markdown
from markdown.extensions import Extension
from markdown.extensions.fenced_code import FencedBlockPreprocessor


class MermaidExtension(Extension):
    def extendMarkdown(self, md: Markdown):
        md.registerExtension(self)
        self.parser = md.parser
        self.md = md

        md.preprocessors.register(MermaidPreprocessor(md, {}), 'mermaid', 30)


class MermaidPreprocessor(FencedBlockPreprocessor):
    __generated_file_contents_cache = {}

    def generate_file_content(self, code):
        if code in self.__generated_file_contents_cache:
            return self.__generated_file_contents_cache[code]

        with TemporaryDirectory() as tempdir:
            puppeteer_config_path = os.path.join(tempdir, "puppeteer-config.json")
            with open(puppeteer_config_path, 'w') as config_file:
                config_file.write(
                    json.dumps(
                        {
                            "args": [
                                "--no-sandbox",
                                "--disable-gpu",
                            ],
                        }
                    )
                )

            input_path = os.path.join(tempdir, "in.mmd")
            with open(input_path, 'w') as input_file:
                input_file.write(code)

            output_path = os.path.join(tempdir, "out.png")
            try:
                check_call(
                    [
                        "mmdc",
                        "-w",
                        "2500",
                        "-s",
                        "2",
                        "-i",
                        input_path,
                        "-o",
                        output_path,
                        "-b",
                        "transparent",
                        "-p",
                        puppeteer_config_path,
                    ],
                    shell=True,
                )

            except CalledProcessError as error:
                raise ValueError("Cannot run mmdc to convert mermaid to image") from error

            with open(output_path, "rb") as out_file:
                self.__generated_file_contents_cache[code] = ''.join(str(base64.b64encode(out_file.read()), 'utf-8').splitlines())

            return self.__generated_file_contents_cache[code]

    def run(self, lines: list[str]) -> list[str]:
        """Match and store Fenced Code Blocks in the `HtmlStash`."""

        text = "\n".join(lines)
        pos = 0
        while 1:
            m = self.FENCED_BLOCK_RE.search(text, pos)
            if m:
                pos = m.end()
                if not m.group('lang') == 'mermaid':
                    continue

                file_content = self.generate_file_content(m.group('code'))
                placeholder = self.md.htmlStash.store(f'<img src="data:image/png;base64,{file_content}">')
                text = f'{text[:m.start()]}\n{placeholder}\n{text[m.end():]}'
                pos = m.start() + len(placeholder)

            else:
                break

        return text.split("\n")
